
import random
from collections import OrderedDict
from typing import IO, Any, Optional, Iterable, Mapping, Dict, \
    NamedTuple, ClassVar, TypeVar, Type

from pymap.mailbox import MailboxSnapshot

from .io import FileWriteable

__all__ = ['Record', 'UidList']

_UDT = TypeVar('_UDT', bound='UidList')


class Record(NamedTuple):
    """Defines a single record read from the UID list file.

    Args:
        uid: The message UID of the record.
        fields: The metadata fields of the record.
        filename: The filename of the record.

    """

    uid: int
    fields: Mapping[str, Any]
    filename: str

    @property
    def key(self) -> str:
        """The :class:`~mailbox.Maildir` key value."""
        return self.filename.split(':', 1)[0]


class UidList(FileWriteable):
    """Maintains the file with UID mapping to maildir files.

    Args:
        base_dir: The directory of the file.
        uid_validity: The UID validity value.
        next_uid: The next assignable message UID value.
        global_uid: The 128-bit global mailbox UID.

    """

    #: The UID list file name, stored in the mailbox directory.
    FILE_NAME: ClassVar[str] = 'dovecot-uidlist'

    #: The UID list lock file, stored adjacent to the UID list file.
    LOCK_FILE: ClassVar[str] = 'dovecot-uidlist.lock'

    def __init__(self, base_dir: str, uid_validity: int,
                 next_uid: int, global_uid: bytes = None) -> None:
        super().__init__()
        self._base_dir = base_dir
        self.uid_validity = uid_validity
        self.next_uid = next_uid
        self.global_uid = global_uid or self._create_guid()
        self._records: Dict[int, Record] = OrderedDict()

    @property
    def records(self) -> Iterable[Record]:
        """The records contained in the UID list file."""
        return self._records.values()

    def get(self, uid: int) -> Record:
        """Get a single record by its UID.

        Args:
            uid: The message UID.

        Raises:
            KeyError: The UID does not exist.

        """
        return self._records[uid]

    def get_all(self, uids: Iterable[int]) -> Mapping[int, Record]:
        """Get records by a set of UIDs.

        Args:
            uids: The message UIDs.

        """
        return {uid: self._records[uid] for uid in uids
                if uid in self._records}

    def set(self, rec: Record) -> None:
        """Add or update the record in the UID list file."""
        self._records[rec.uid] = rec

    def remove(self, uid: int) -> None:
        """Remove the record from the UID list file.

        Raises:
            KeyError: The UID does not exist.

        """
        del self._records[uid]

    @classmethod
    def _build_line(cls, rec: Record) -> str:
        parts = ['%d' % rec.uid]
        for key, val in sorted(rec.fields.items()):
            parts.append(' ')
            parts.append(key[0:1])
            parts.append(str(val))
        parts.append(' :')
        parts.append(rec.filename)
        parts.append('\r\n')
        return ''.join(parts)

    @classmethod
    def _read_line(cls, line: str) -> Record:
        before, filename = line.split(':', 1)
        fields: Dict[str, str] = {}
        data = before.split(' ')
        num = int(data[0])
        for col in data[1:]:
            if col:
                fields[col[0]] = col[1:]
        return Record(num, fields, filename.rstrip())

    @classmethod
    def _read_header(cls: Type[_UDT], base_dir: str, line: str) -> _UDT:
        data = line.split()
        if data[0] != '3':
            raise ValueError(line)
        uid_validity: Optional[int] = None
        next_uid: Optional[int] = None
        global_uid: Optional[bytes] = None
        for field in data[1:]:
            if field[0] == 'V':
                uid_validity = int(field[1:])
            elif field[0] == 'N':
                next_uid = int(field[1:])
            elif field[0] == 'G':
                global_uid = field[1:].encode('ascii')
        if uid_validity is None or next_uid is None or global_uid is None:
            raise ValueError(line)
        return cls(base_dir, uid_validity, next_uid, global_uid)

    @classmethod
    def _create_guid(cls) -> bytes:
        return b'%032x' % random.getrandbits(128)

    def _build_header(self) -> str:
        global_uid = self.global_uid.decode('ascii')
        return ''.join(['3 V', str(self.uid_validity),
                        ' N', str(self.next_uid),
                        ' G', global_uid, '\r\n'])

    @classmethod
    def get_file(cls) -> str:
        return cls.FILE_NAME

    @classmethod
    def get_lock(cls) -> str:
        return cls.LOCK_FILE

    def get_dir(self) -> str:
        return self._base_dir

    @classmethod
    def get_default(cls: Type[_UDT], base_dir: str) -> _UDT:
        return cls(base_dir, MailboxSnapshot.new_uid_validity(), 1)

    def write(self, fp: IO[str]) -> None:
        fp.write(self._build_header())
        for rec in self.records:
            fp.write(self._build_line(rec))

    @classmethod
    def open(cls: Type[_UDT], base_dir: str, fp: IO[str]) -> _UDT:
        header = fp.readline()
        ret = cls._read_header(base_dir, header)
        return ret

    def read(self, fp: IO[str]) -> None:
        for line in fp:
            self.set(self._read_line(line))




from abc import abstractmethod, ABCMeta
from argparse import Namespace
from typing import TextIO
from typing_extensions import Final

from ..grpc.admin_grpc import AdminStub

__all__ = ['ClientCommand']


class ClientCommand(metaclass=ABCMeta):
    """Interface for client command implementations.

    Args:
        stub: The GRPC stub for executing commands.
        args: The command line arguments.

    """

    def __init__(self, stub: AdminStub, args: Namespace) -> None:
        super().__init__()
        self.stub: Final = stub
        self.args: Final = args

    @classmethod
    @abstractmethod
    def add_subparser(cls, name: str, subparsers) -> None:
        """Add the command-line argument subparser for the command.

        Args:
            name: The name to use for the subparser.
            subparsers: The special action object as returned by
                :meth:`~argparse.ArgumentParser.add_subparsers`.

        """
        ...

    @abstractmethod
    async def run(self, fileobj: TextIO) -> int:
        """Run the command and return the exit code.

        Args:
            fileobj: The file object to print the output to.

        """
        ...



"""IMAP server with pluggable Python backends."""

import argparse
import asyncio
import logging
import logging.config
import os
from argparse import ArgumentParser, Namespace, ArgumentTypeError
from typing import Any, Type, Sequence, Mapping

from pkg_resources import iter_entry_points, DistributionNotFound

from . import __version__
from .interfaces.backend import BackendInterface, ServiceInterface

try:
    import systemd.daemon  # type: ignore
except ImportError:
    def notify_ready() -> None:
        pass
else:
    def notify_ready() -> None:
        systemd.daemon.notify('READY=1')

_Backends = Mapping[str, Type[BackendInterface]]
_Services = Mapping[str, Type[ServiceInterface]]


def main() -> None:
    parser = ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--debug', action='store_true',
                        help='increase printed output for debugging')
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('--set-uid', metavar='USER', type=_get_pwd,
                        help='drop privileges to user name or uid')
    parser.add_argument('--set-gid', metavar='GROUP', type=_get_grp,
                        help='drop privileges to group name or gid')
    parser.add_argument('--logging-cfg', metavar='PATH',
                        help='config file for logging')
    parser.add_argument('--no-service', dest='skip_services', action='append',
                        metavar='NAME', help='do not run the given service')
    subparsers = parser.add_subparsers(dest='backend',
                                       help='which pymap backend to use')

    backends: _Backends = _load_entry_points('pymap.backend')
    services: _Services = _load_entry_points('pymap.service')

    for backend_name, backend_cls in backends.items():
        backend_cls.add_subparser(backend_name, subparsers)
    for service_cls in services.values():
        service_cls.add_arguments(parser)
    parser.set_defaults(skip_services=[])
    args = parser.parse_args()

    if args.logging_cfg:
        logging.config.fileConfig(args.logging_config)
    else:
        logging.basicConfig(level=logging.WARNING)
    if args.debug:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    if not args.backend:
        parser.error('Expected backend name')
    backend_type = backends[args.backend]

    service_types = [service for name, service in services.items()
                     if name not in args.skip_services]

    try:
        return asyncio.run(run(args, backend_type, service_types), debug=False)
    except KeyboardInterrupt:
        pass


async def run(args: Namespace, backend_type: Type[BackendInterface],
              service_types: Sequence[Type[ServiceInterface]]) -> None:
    backend, config = await backend_type.init(args)
    config.apply_context()
    services = [await service.start(backend, config)
                for service in service_types]

    _drop_privileges(args)
    notify_ready()
    await asyncio.gather(*[service.task for service in services])


def _load_entry_points(group: str) -> Mapping[str, Type[Any]]:
    ret = {}
    for entry_point in iter_entry_points(group):
        try:
            cls = entry_point.load()
        except DistributionNotFound:
            pass  # optional dependencies not installed
        else:
            ret[entry_point.name] = cls
    return ret


def _get_pwd(setuid: str) -> int:
    from pwd import getpwnam, getpwuid
    try:
        try:
            uid = int(setuid)
        except ValueError:
            entry = getpwnam(setuid)
        else:
            entry = getpwuid(uid)
    except KeyError as exc:
        raise ArgumentTypeError(f'Invalid user: {setuid}') from exc
    else:
        return entry.pw_uid


def _get_grp(setgid: str) -> int:
    from grp import getgrnam, getgrgid
    try:
        try:
            gid = int(setgid)
        except ValueError:
            entry = getgrnam(setgid)
        else:
            entry = getgrgid(gid)
    except KeyError as exc:
        raise ArgumentTypeError(f'Invalid group: {setgid}') from exc
    else:
        return entry.gr_gid


def _drop_privileges(args: Namespace) -> None:
    if args.set_gid is not None:
        os.setgid(args.set_gid)
    if args.set_uid is not None:
        os.setuid(args.set_uid)




from datetime import datetime, timezone
from typing_extensions import Final

from pymap.exceptions import ResponseError
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.message import AppendMessage
from pymap.interfaces.session import SessionInterface
from pymap.parsing.specials import Flag, ExtensionOptions
from pysasl.external import ExternalResult

from .grpc.admin_grpc import AdminBase

__all__ = ['AdminHandlers']


class AdminHandlers(AdminBase):
    """The GRPC handlers, executed when an admin request is received. Each
    handler should receive a request, take action, and send the response.

    See Also:
        :class:`grpclib.server.Server`

    """

    def __init__(self, backend: BackendInterface) -> None:
        super().__init__()
        self.config: Final = backend.config
        self.login: Final = backend.login

    async def _login_as(self, user: str) -> SessionInterface:
        creds = ExternalResult(user)
        return await self.login(creds, self.config)

    @property
    def _with_filter(self) -> bool:
        return not self.config.args.no_filter

    async def Append(self, stream) -> None:
        """Append a message directly to a user's mailbox.

        If the backend session defines a
        :attr:`~pymap.interfaces.session.SessionInterface.filter_set`, the
        active filter implementation will be applied to the appended message,
        such that the message may be discarded, modified, or placed into a
        specific mailbox.

        For example, using the CLI client::

            $ cat message.txt | pymap-admin append user@example.com

        See ``pymap-admin append --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        from .grpc.admin_pb2 import AppendRequest, AppendResponse, \
            ERROR_RESPONSE
        request: AppendRequest = await stream.recv_message()
        mailbox = request.mailbox or 'INBOX'
        flag_set = frozenset(Flag(flag) for flag in request.flags)
        when = datetime.fromtimestamp(request.when, timezone.utc)
        append_msg = AppendMessage(request.data, flag_set, when,
                                   ExtensionOptions.empty())
        try:
            session = await self._login_as(request.user)
            if self._with_filter and session.filter_set is not None:
                filter_value = await session.filter_set.get_active()
                if filter_value is not None:
                    filter_ = await session.filter_set.compile(filter_value)
                    new_mailbox, append_msg = await filter_.apply(
                        request.sender, request.recipient, mailbox, append_msg)
                    if new_mailbox is None:
                        await stream.send_message(AppendResponse())
                        return
                    else:
                        mailbox = new_mailbox
            append_uid, _ = await session.append_messages(
                mailbox, [append_msg])
        except ResponseError as exc:
            resp = AppendResponse(result=ERROR_RESPONSE,
                                  error_type=type(exc).__name__,
                                  error_response=bytes(exc.get_response(b'.')),
                                  mailbox=mailbox)
            await stream.send_message(resp)
        else:
            validity = append_uid.validity
            uid = next(iter(append_uid.uids))
            resp = AppendResponse(mailbox=mailbox, validity=validity, uid=uid)
            await stream.send_message(resp)




from typing import Union, Tuple, Iterable, Sequence, List
from typing_extensions import Final

from .bytes import MaybeBytes, Writeable
from .interfaces.message import MessageInterface
from .parsing.primitives import Number, ListP, LiteralString
from .parsing.specials import DateTime, FetchAttribute
from .selected import SelectedMailbox

__all__ = ['NotFetchable', 'MessageAttributes']

_Partial = Union[None, Tuple[int, int], Tuple[int]]


def _not_expunged(orig):
    def deco(self: 'MessageAttributes', attr: FetchAttribute,
             *args, **kwargs) -> MaybeBytes:
        if self.message.expunged:
            raise NotFetchable(attr)
        else:
            return orig(self, attr, *args, **kwargs)
    return deco


class NotFetchable(Exception):
    """Raised when a message cannot provide a
    :class:`~pymap.parsing.specials.fetchattr.FetchAttribute` for some reason,
    e.g. because it has been expunged.

    """

    def __init__(self, attr: FetchAttribute) -> None:
        attr_name = attr.value.decode('ascii')
        super().__init__(f'{attr_name} is not fetchable')
        self._attr = attr

    @property
    def attr(self) -> FetchAttribute:
        """The attribute that could not be fetched."""
        return self._attr


class MessageAttributes:
    """Defines a re-usable object to query a
    :class:`~pymap.parsing.specials.fetchattr.FetchAttribute` from a message.

    Args:
        message: The message object.
        selected: The selected mailbox.

    """

    def __init__(self, message: MessageInterface,
                 selected: SelectedMailbox) -> None:
        super().__init__()
        self.message: Final = message
        self.selected: Final = selected

    def get_all(self, attrs: Iterable[FetchAttribute]) \
            -> Sequence[Tuple[FetchAttribute, MaybeBytes]]:
        """Return a list of tuples containing the attribute iself and the bytes
        representation of that attribute from the message.

        Args:
            attrs: The fetch attributes.

        """
        ret: List[Tuple[FetchAttribute, MaybeBytes]] = []
        for attr in attrs:
            try:
                ret.append((attr.for_response, self.get(attr)))
            except NotFetchable:
                pass
        return ret

    def get(self, attr: FetchAttribute) -> MaybeBytes:
        """Return the bytes representation of the given message attribue.

        Args:
            attr: The fetch attribute.

        Raises:
            :class:`NotFetchable`

        """
        attr_name = attr.value.decode('ascii')
        method = getattr(self, '_get_' + attr_name.replace('.', '_'))
        return method(attr)

    def _get_data(self, section: FetchAttribute.Section, partial: _Partial, *,
                  binary: bool = False) -> Writeable:
        msg = self.message
        specifier = section.specifier
        parts = section.parts
        headers = section.headers
        if specifier is None:
            data = msg.get_body(parts, binary)
        elif specifier == b'MIME' and parts is not None:
            data = msg.get_headers(parts)
        elif specifier == b'TEXT':
            data = msg.get_message_text(parts)
        elif specifier == b'HEADER':
            data = msg.get_message_headers(parts)
        elif specifier == b'HEADER.FIELDS':
            data = msg.get_message_headers(parts, headers)
        elif specifier == b'HEADER.FIELDS.NOT':
            data = msg.get_message_headers(parts, headers, True)
        else:
            raise RuntimeError()  # Should not happen.
        return self._get_partial(data, partial)

    def _get_partial(self, data: Writeable, partial: _Partial) -> Writeable:
        if partial is None:
            return data
        full = bytes(data)
        start = int(partial[0])
        try:
            end = int(partial[1])
        except IndexError:
            end = len(full)
        return Writeable.wrap(full[start:end])

    def _get_UID(self, attr: FetchAttribute) -> MaybeBytes:
        return Number(self.message.uid)

    def _get_FLAGS(self, attr: FetchAttribute) -> MaybeBytes:
        session_flags = self.selected.session_flags
        flag_set = self.message.get_flags(session_flags)
        return ListP(flag_set, sort=True)

    def _get_INTERNALDATE(self, attr: FetchAttribute) -> MaybeBytes:
        return DateTime(self.message.internal_date)

    @_not_expunged
    def _get_ENVELOPE(self, attr: FetchAttribute) -> MaybeBytes:
        return self.message.get_envelope_structure()

    @_not_expunged
    def _get_BODYSTRUCTURE(self, attr: FetchAttribute) -> MaybeBytes:
        return self.message.get_body_structure().extended

    @_not_expunged
    def _get_BODY(self, attr: FetchAttribute) -> MaybeBytes:
        if attr.section is None:
            return self.message.get_body_structure()
        return LiteralString(self._get_data(attr.section, attr.partial))

    @_not_expunged
    def _get_BODY_PEEK(self, attr: FetchAttribute) -> MaybeBytes:
        return self._get_BODY(attr)

    @_not_expunged
    def _get_RFC822(self, attr: FetchAttribute) -> MaybeBytes:
        return self._get_BODY(attr)

    @_not_expunged
    def _get_RFC822_HEADER(self, attr: FetchAttribute) -> MaybeBytes:
        return self._get_BODY(attr)

    @_not_expunged
    def _get_RFC822_TEXT(self, attr: FetchAttribute) -> MaybeBytes:
        return self._get_BODY(attr)

    @_not_expunged
    def _get_RFC822_SIZE(self, attr: FetchAttribute) -> MaybeBytes:
        return Number(self.message.get_size())

    @_not_expunged
    def _get_BINARY(self, attr: FetchAttribute) -> MaybeBytes:
        if attr.section is None:
            raise RuntimeError()  # should not happen.
        data = self._get_data(attr.section, attr.partial, binary=True)
        return LiteralString(data, True)

    @_not_expunged
    def _get_BINARY_PEEK(self, attr: FetchAttribute) -> MaybeBytes:
        return self._get_BINARY(attr)

    @_not_expunged
    def _get_BINARY_SIZE(self, attr: FetchAttribute) -> MaybeBytes:
        if attr.section is None:
            raise RuntimeError()  # should not happen.
        data = self._get_data(attr.section, attr.partial, binary=True)
        return Number(len(data))




import enum
from io import BytesIO
from typing import Optional, Mapping, Sequence

from pymap.bytes import BytesFormat, MaybeBytes, WriteStream, Writeable
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import String, QuotedString, LiteralString

__all__ = ['Condition', 'Response', 'BadCommandResponse',
           'CapabilitiesResponse', 'GetScriptResponse', 'ListScriptsResponse']

_Capabilities = Mapping[bytes, Optional[MaybeBytes]]


class Condition(enum.Enum):
    OK = enum.auto()
    NO = enum.auto()
    BYE = enum.auto()

    def __bytes__(self) -> bytes:
        return bytes(str(self.name), 'ascii')


class Response(Writeable):

    def __init__(self, condition: Condition, *, code: MaybeBytes = None,
                 text: str = None) -> None:
        super().__init__()
        self.condition = condition
        self.code = code
        self.text = text
        self._raw: Optional[bytes] = None

    @property
    def is_bye(self) -> bool:
        return self.condition == Condition.BYE

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self.condition))
        if self.code is not None:
            writer.write(b' (%b)' % bytes(self.code))
        if self.text is not None:
            text_bytes = self.text.rstrip('\r\n').encode('utf-8', 'replace')
            text_str = String.build(text_bytes)
            writer.write(b' ')
            text_str.write(writer)
        writer.write(b'\r\n')

    def __bytes__(self) -> bytes:
        if self._raw is None:
            out = BytesIO()
            self.write(out)
            self._raw = out.getvalue()
        return self._raw

    def __repr__(self) -> str:
        raw = bytes(self).decode('utf-8')
        return f'<Response {repr(raw)}>'


class BadCommandResponse(Response):

    def __init__(self, exc: NotParseable) -> None:
        super().__init__(Condition.NO, text='Bad command: %s' % exc)
        self.exc = exc


class NoOpResponse(Response):

    def __init__(self, tag: Optional[bytes]) -> None:
        if tag is None:
            code: Optional[bytes] = None
        else:
            code = BytesFormat(b'TAG %b') % String.build(tag)
        super().__init__(Condition.OK, code=code)


class CapabilitiesResponse(Response):

    def __init__(self, capabilities: _Capabilities, *,
                 code: MaybeBytes = None) -> None:
        super().__init__(Condition.OK, code=code)
        self.capabilities = capabilities

    def write(self, writer: WriteStream) -> None:
        for cap_name, cap_val in self.capabilities.items():
            quoted_name = QuotedString(cap_name)
            if cap_val is None:
                line = BytesFormat(b'%b\r\n') % (quoted_name, )
            else:
                quoted_val = QuotedString(bytes(cap_val))
                line = BytesFormat(b'%b %b\r\n') % (quoted_name, quoted_val)
            writer.write(line)
        super().write(writer)


class GetScriptResponse(Response):

    def __init__(self, script_data: bytes) -> None:
        super().__init__(Condition.OK)
        self.script_data = script_data

    def write(self, writer: WriteStream) -> None:
        data_str = LiteralString(self.script_data)
        data_str.write(writer)
        writer.write(b'\r\n')
        super().write(writer)


class ListScriptsResponse(Response):

    def __init__(self, active: Optional[str], names: Sequence[str]) -> None:
        super().__init__(Condition.OK)
        self.active = active
        self.names = names

    def write(self, writer: WriteStream) -> None:
        for name in self.names:
            name_str = String.build(name)
            name_str.write(writer)
            if self.active and name == self.active:
                writer.write(b' ACTIVE\r\n')
            else:
                writer.write(b'\r\n')
        super().write(writer)



"""Package of special data objects, usually composed of one of more
primitives.

"""

__all__ = ['AString', 'DateTime', 'FetchAttribute', 'FetchRequirement',
           'Flag', 'Mailbox', 'SearchKey', 'SequenceSet', 'StatusAttribute',
           'Tag', 'ExtensionOption', 'ExtensionOptions']

from .astring import AString
from .datetime_ import DateTime
from .fetchattr import FetchAttribute, FetchRequirement
from .flag import Flag
from .mailbox import Mailbox
from .searchkey import SearchKey
from .sequenceset import SequenceSet
from .statusattr import StatusAttribute
from .options import ExtensionOption, ExtensionOptions
from .tag import Tag




import ipaddress
import socket as _socket
from abc import abstractmethod, ABCMeta
from asyncio import BaseTransport, StreamWriter
from typing import Any, Union, Optional, Tuple, Sequence, Mapping, List

try:
    import systemd.daemon  # type: ignore
except ImportError as exc:
    systemd_import_exc: Optional[ImportError] = exc
else:
    systemd_import_exc = None

__all__ = ['InheritedSockets', 'SocketInfo']

_Transport = Union[BaseTransport, StreamWriter]
_PeerName = Union[Tuple[str, int], Tuple[str, int, int, int], str]
_PeerCert = Mapping[str, Any]  # {'issuer': ..., ...}


class InheritedSockets(metaclass=ABCMeta):
    """Abstracts the ability to retrieve inherited sockets from the calling
    process, usually a service management framework.

    """

    @abstractmethod
    def get(self) -> Sequence[_socket.socket]:
        """Return the sockets inherited by the process."""
        ...

    @classmethod
    def of(cls, service_manager: str) -> 'InheritedSockets':
        """Return the inherited sockets for the given service manager.

        Note:
            Only ``'systemd'`` is implemented at this time.

        Args:
            service_manager: The service manager name.

        """
        if service_manager == 'systemd':
            return cls.for_systemd()
        else:
            raise KeyError(service_manager)

    @classmethod
    def for_systemd(cls) -> 'InheritedSockets':
        """Return the inherited sockets for `systemd`_. The `python-systemd`_
        library must be installed.

        See Also:
            `systemd.socket`_, `sd_listen_fds`_

        .. _systemd: https://freedesktop.org/wiki/Software/systemd/
        .. _systemd.socket: https://www.freedesktop.org/software/systemd/man/systemd.socket.html
        .. _sd_listen_fds: https://www.freedesktop.org/software/systemd/man/sd_listen_fds.html
        .. _python-systemd: https://github.com/systemd/python-systemd

        Raises:
            :exc:`NotImplementedError`

        """  # noqa: E501
        if systemd_import_exc:
            raise systemd_import_exc
        return _SystemdSockets()


class _SystemdSockets(InheritedSockets):

    def __init__(self) -> None:
        super().__init__()
        fds: Sequence[int] = systemd.daemon.listen_fds()
        sockets: List[_socket.socket] = []
        for fd in fds:
            family = self._get_family(fd)
            sock = _socket.fromfd(fd, family, _socket.SOCK_STREAM)
            sockets.append(sock)
        self._sockets = sockets

    def get(self) -> Sequence[_socket.socket]:
        return self._sockets

    @classmethod
    def _get_family(cls, fd: int) -> int:
        if systemd.daemon.is_socket(fd, _socket.AF_UNIX):
            return _socket.AF_UNIX
        elif systemd.daemon.is_socket(fd, _socket.AF_INET6):
            return _socket.AF_INET6
        else:
            return _socket.AF_INET


class SocketInfo:
    """Information about a connected socket, which may be useful for
    server-side logic such as authentication and authorization.

    Attribute access is passed directly into
    :meth:`~asyncio.BaseTransport.get_extra_info`, decorated in some cases with
    type hints and checks.

    See Also:
        :meth:`~asyncio.BaseTransport.get_extra_info`

    """

    __slots__ = ['_transport']

    def __init__(self, transport: _Transport) -> None:
        super().__init__()
        self._transport = transport

    @property
    def socket(self) -> _socket.socket:
        sock = self._transport.get_extra_info('socket')
        if sock is None:
            raise ValueError('socket')
        return sock

    @property
    def peername(self) -> _PeerName:
        peername = self._transport.get_extra_info('peername')
        if peername is None:
            raise ValueError('peername')
        return peername

    @property
    def peercert(self) -> Optional[_PeerCert]:
        return self._transport.get_extra_info('peercert')

    @property
    def from_localhost(self) -> bool:
        """True if :attr:`.peername` is a connection from a ``localhost``
        address.

        """
        sock_family = self.socket.family
        if sock_family == _socket.AF_UNIX:
            return True
        elif sock_family not in (_socket.AF_INET, _socket.AF_INET6):
            return False
        sock_address, *_ = self.peername
        ip = ipaddress.ip_address(sock_address)
        if ip.version == 6 and ip.ipv4_mapped is not None:
            ip = ipaddress.ip_address(ip.ipv4_mapped)
        return ip.is_loopback

    def __getattr__(self, name: str) -> Any:
        return self._transport.get_extra_info(name)

    def __str__(self) -> str:
        return '<SocketInfo peername=%r sockname=%r peercert=%r>' \
            % (self.peername, self.sockname, self.peercert)

    def __bytes__(self) -> bytes:
        return bytes(str(self), 'utf-8', 'replace')



"""Implements the modified UTF-7 specification used for encoding and decoding
mailbox names in IMAP.

See Also:
    `RFC 3501 5.1.3 <https://tools.ietf.org/html/rfc3501#section-5.1.3>`_

"""

__all__ = ['modutf7_encode', 'modutf7_decode']


def _modified_b64encode(src: str) -> bytes:
    # Inspired by Twisted Python's implementation:
    #   https://twistedmatrix.com/trac/browser/trunk/LICENSE
    src_utf7 = src.encode('utf-7')
    return src_utf7[1:-1].replace(b'/', b',')


def _modified_b64decode(src: bytes) -> str:
    # Inspired by Twisted Python's implementation:
    #   https://twistedmatrix.com/trac/browser/trunk/LICENSE
    src_utf7 = b'+%b-' % src.replace(b',', b'/')
    return src_utf7.decode('utf-7')


def modutf7_encode(data: str) -> bytes:
    """Encode the string using modified UTF-7.

    Args:
        data: The input string to encode.

    """
    ret = bytearray()
    is_usascii = True
    encode_start = None
    for i, symbol in enumerate(data):
        charpoint = ord(symbol)
        if is_usascii:
            if charpoint == 0x26:
                ret.extend(b'&-')
            elif 0x20 <= charpoint <= 0x7e:
                ret.append(charpoint)
            else:
                encode_start = i
                is_usascii = False
        else:
            if 0x20 <= charpoint <= 0x7e:
                to_encode = data[encode_start:i]
                encoded = _modified_b64encode(to_encode)
                ret.append(0x26)
                ret.extend(encoded)
                ret.extend((0x2d, charpoint))
                is_usascii = True
    if not is_usascii:
        to_encode = data[encode_start:]
        encoded = _modified_b64encode(to_encode)
        ret.append(0x26)
        ret.extend(encoded)
        ret.append(0x2d)
    return bytes(ret)


def modutf7_decode(data: bytes) -> str:
    """Decode the bytestring using modified UTF-7.

    Args:
        data: The encoded bytestring to decode.

    """
    parts = []
    is_usascii = True
    buf = memoryview(data)
    while buf:
        byte = buf[0]
        if is_usascii:
            if buf[0:2] == b'&-':
                parts.append('&')
                buf = buf[2:]
            elif byte == 0x26:
                is_usascii = False
                buf = buf[1:]
            else:
                parts.append(chr(byte))
                buf = buf[1:]
        else:
            for i, byte in enumerate(buf):
                if byte == 0x2d:
                    to_decode = buf[:i].tobytes()
                    decoded = _modified_b64decode(to_decode)
                    parts.append(decoded)
                    buf = buf[i + 1:]
                    is_usascii = True
                    break
    if not is_usascii:
        to_decode = buf.tobytes()
        decoded = _modified_b64decode(to_decode)
        parts.append(decoded)
    return ''.join(parts)
